// URL Lock

//
// This file contains the implementation of the URL Lock Library interfaces
//

// XPCOM structure adapted from http://developer.mozilla.org/en/docs/How_to_Build_an_XPCOM_Component_in_Javascript

const IURLLockEngine = Components.interfaces.IURLLockEngine;
const nsISupports = Components.interfaces.nsISupports;
const CLASS_ID = Components.ID("{ab432666-2f11-4db5-89b9-16cbadb3e5f6}");
const CLASS_NAME = "URL Lock Engine";
const CONTRACT_ID = "@ieurllock.sourceforge.net/engine;1";

/**
 * Create a new protocols container object, initializing all
 * properties to false in the process
 */
function LockProtocols() {
    this.wrappedJSObject = this;
    this.file = false;
    this.ftp = false;
    this.http = false;
    this.https = false;
}

LockProtocols.prototype = {

    /** When true, file:// applies to the rule */
    get file() {
        return this.file;
    },
    
    /** When true, file:// applies to the rule */
    set file(file) {
        this.file = file;
    },
    
    /** When true, FTP applies to the rule */
    get ftp() {
        return this.ftp;
    },
    
    /** When true, FTP applies to the rule */
    set ftp(ftp) {
        this.ftp = ftp;
    },
    
    /** When true, HTTP applies to the rule */
    get http() {
        return this.http;
    },
    
    /** When true, HTTP applies to the rule */
    set http(http) {
        this.http = http;
    },
    
    /** When true, HTTPS applies to the rule */
    get https() {
        return this.https;
    },

    /** When true, HTTPS applies to the rule */
    set https(https) {
        this.https = https;
    },

    /**
     * Returns the requested interface for this object instance
     */
    QueryInterface: function(aIID)
    {
        if (!aIID.equals(Components.interfaces.IURLLockProtocols) &&
            !aIID.equals(nsISupports))
            throw Components.results.NS_ERROR_NO_INTERFACE;
        return this;
    }
}

/**
 * Creates a new lock browser link for the given location and last
 * access time
 */
function LockBrowserLink(location, lastAccessTime) {
    this.wrappedJSObject = this;
    this.lastAccessTime = lastAccessTime;
    this.location = location;
}

LockBrowserLink.prototype = {

    /**
     * lastAccessTime: date: The last-access time of this URL, which is null if not known
     */
    get lastAccessTime() {
        return this.lastAccessTime;
    },
    
    /**
     * lastAccessTime: date: The last-access time of this URL, which is null if not known
     */
    set lastAccessTime(lastAccessTime) {
        this.lastAccessTime = lastAccessTime;
    },
    
    /**
     * location: String: The location URL or path
     */
    get location() {
        return this.location;
    },

    /**
     * location: String: The location URL or path
     */
    set location(location) {
        this.location = location;
    },
    
    /**
     * Returns the requested interface for this object instance
     */
    QueryInterface: function(aIID)
    {
        if (!aIID.equals(Components.interfaces.IURLLockBrowserLink) &&    
            !aIID.equals(nsISupports))
            throw Components.results.NS_ERROR_NO_INTERFACE;
        return this;
    }
}

/**
 * Creates a list of lock browser links using the given array
 */
function LockBrowserLinkList(array) {
    this.array = array;
}

LockBrowserLinkList.prototype = {

    item: function(index) {
        return this.array[index];
    },
    
    get length() {
        return this.array.length;
    }
}

/**
 * Creates a browser links grouping object that is used to get
 * histories and bookmarks for a specific user and web browser
 */
function LockBrowserLinks(user) {
    this.wrappedJSObject = this;

    // Set the user that we will get browser links for
    this.user = user;
}

LockBrowserLinks.prototype = {

    /**
     * getInternetExplorerBookmarks(): LockBrowserLink[]: Returns the bookmarks defined in
     * Internet Explorer for the user interface
     *
     * Owner: Annie Zhao
     */
    getInternetExplorerBookmarks: function() {
        
        // Get the native service
        var lockNative = Components.classes["@ieurllock.sourceforge.net/native;1"]
                         .getService(Components.interfaces.IURLLockNative);
        
        // Ask the native component for the IE favorites path
        var favoritesPath = lockNative.getInternetExplorerFavoritesPath();
        
        // TODO: Read in the favorites
        
        // Return the favorites
        return new LockBrowserLinkList(new Array());
    },
    
    /**
     * getInternetExplorerHistory(): LockBrowserLink[]: Returns the browser history in
     * Internet Explorer for the user interface
     *
     * Owner: Steven Lawrance
     */
    getInternetExplorerHistory: function() {

        // Get the native service
        var lockNative = Components.classes["@ieurllock.sourceforge.net/native;1"]
                         .getService(Components.interfaces.IURLLockNative);
        
        // Return the history (current user only in this version)
        return getInternetExplorerHistory();
    },
    
    /**
     * Returns the requested interface for this object instance
     */
    QueryInterface: function(aIID)
    {
        if (!aIID.equals(Components.interfaces.IURLLockBrowserLinks) &&    
            !aIID.equals(nsISupports))
            throw Components.results.NS_ERROR_NO_INTERFACE;
        return this;
    }
}

/**
 * Creates a new user object using the given user ID
 */
function LockUser(configuration, userID) {
    this.wrappedJSObject = this;

    // Set our configuration reference
    this.configuration = configuration;
    
    // Set the opaque userID reference
    this.userID = userID;
    
    // Create our properties object
    this.properties = new Object();
    
    // Set our oldProperties to null to indicate that we have no changes
    this.oldProperties = null;

    // Get the native service
    var lockNative = Components.classes["@ieurllock.sourceforge.net/native;1"]
                     .getService(Components.interfaces.IURLLockNative);
                         
    // Get our user's name
    this.properties.userName = lockNative.getUserName(this.userID);
    
    // Get the user's full name
    this.properties.fullName = lockNative.getUserFullName(this.userID);
    if (this.properties.fullName == "") {
        this.properties.fullName = this.properties.userName;
    }
    
    // Get the user's administrator privilege status
    this.properties.administrator = lockNative.isAdministrator(this.userID);
    
    // TODO: Load the other settings from the registry
}

LockUser.prototype = {
    
    /**
     * fieldChanged(fieldName: String, fieldValue: String): void: A field in this user changed
     *
     * Owner: Steven Lawrance
     */
    fieldChanged: function(fieldName, fieldValue) {
    },
    
    /**
     * getAllowAllByDefault(): boolean: Returns true if all locations are enabled by
     * default or false if all locations are blocked by default. This substantially
     * affects the operation of the rule list. If this function returns true, rules
     * in the rule list will block matching locations. If this function returns false,
     * rules in the rule list will allow matching locations
     *
     * Owner: Steven Lawrance
     */
    getAllowAllByDefault: function() {
        return this.properties.allowAllByDefault;
    },
    
    /**
     * getBlockLocation(): String: Returns the location that the URL Lock will direct the
     * browser to when sites are blocked when the showDialogOnBlock property is false
     *
     * Owner: Steven Lawrance
     */
    getBlockLocation: function() {
        return this.properties.blockLocation;
    },
    
    /**
     * getBrowserLinks(): LockBrowserLinks: Returns the browser links object for this user,
     * which provides access to bookmarks and histories
     *
     * Owner: Steven Lawrance
     */
    getBrowserLinks: function() {
        return new LockBrowserLinks(this);
    },
    
    /**
     * getFullName(): String: Returns the user’s full name
     *
     * Owner: Steven Lawrance
     */
    getFullName: function() {
        return this.properties.fullName;
    },
    
    /**
     * getLockEnabled(): boolean: Returns true if the lock is enabled for this user of false
     * if not
     *
     * Owner: Steven Lawrance
     */
    getLockEnabled: function() {
        return this.properties.lockEnabled;
    },
    
    /**
     * getPreventLockDisable(): boolean: Returns true if the registry permissions are set
     * such that this user cannot disable the URL Lock. Otherwise, this returns false
     *
     * Owner: Steven Lawrance
     */
    getPreventLockDisable: function() {
        return this.properties.preventLockDisable;
    },
    
    /**
     * getProcessAllLocations(): boolean: Returns true if all locations, including drive
     * letters, WebDAV shares, UNC paths, “My Computer,” “Control Panel,” and other such
     * locations, are processed by the URL Lock. Note that this property only affects the
     * Internet Explorer version of the URL Lock and has no effect in the Firefox version.
     * When the return value is false, only web and FTP locations are processed by the
     * URL Lock
     *
     * Owner: Steven Lawrance
     */
    getProcessAllLocations: function() {
        return this.properties.processAllLocations;
    },
    
    /**
     * getShowDialogOnBlock(): boolean: Returns true if the URL Lock shows a dialog box
     * when a site is blocked instead of a block page. Returns false if a block page is
     * shown when a site is blocked, which is defined by the blockLocation property
     *
     * Owner: Steven Lawrance
     */
    getShowDialogOnBlock: function() {
        return this.properties.showDialogOnBlock;
    },
    
    /**
     * getUserName(): String: Returns the user’s system name
     *
     * Owner: Steven Lawrance
     */
    getUserName: function() {
        return this.properties.userName;
    },
    
    /**
     * isAdministrator(): boolean: Returns true if this user is an administrator or false
     * if not
     *
     * Owner: Steven Lawrance
     */
    isAdministrator: function(){
        return this.properties.administrator;
    },
    
    /**
     * isChanged(): boolean: Returns true if at least one property of this user has
     * changed since the last save or load
     *
     * Owner: Steven Lawrance
     */
    isChanged: function() {
    
        // If we have an old properties, then we are changed
        return this.oldProperties != null;
    },
    
    /**
     * Prepares to change the properties by copying the existing properties
     * into the old configuration object. This method is private
     */
    prepareChange: function() {
    
        // If we already have old properties, then return now
        if (this.oldProperties != null) {
            return;
        }
        
        // Create the old properties and copy them
        this.oldProperties = new Object();
        for (i in this.properties) {
            this.oldProperties[i] = this.properties[i];
        }
    },
    
    /**
     * revert(): void: Reverts any changes to the currently-saved state
     */
    revert: function() {
        
        // Forget the current properties
        var revertedProperties = this.properties;
        this.properties = this.oldProperties;
        this.oldProperties = null;
        
        // Generate a change even for every property that differs
        for (i in this.properties) {
            if (this.properties[i] != revertedProperties[i]) {
                this.fieldChanged(i, this.properties[i]);
            }
        }
    },
    
    /**
     * save(): void: Saves any unsaved changes to this user
     *
     * Owner: Steven Lawrance
     */
    save: function() {
        // TODO: Implement
        
        // If we don't have old properties, then there is nothing to save
        if (this.oldProperties == null) {
            return;
        }
        
        // The old properties are no longer relevant
        this.oldProperties = null;
        
        // Update the users list in our configuration
        this.configuration.updateUser(this);
    },
    
    /**
     * setAllowAllByDefault(allow: boolean): void: Sets whether or not locations are
     * allowed by default. This substantially affects the set to false, rules in the
     * rule list will allow matching locations.
     *
     * Owner: Steven Lawrance
     */
    setAllowAllByDefault: function(allow) {
        this.prepareChange();
        
        // Set and notify listeners
        this.properties.allowAllByDefault = allow;
        this.fieldChanged("allowAllByDefault", this.properties.allowAllByDefault);
    },
    
    /**
     * setBlockLocation(blockLocation: String): void: Sets the location that the URL Lock
     * will direct the browser to when sites are blocked when the showDialogOnBlock
     * property is false
     *
     * Owner: Steven Lawrance
     */
    setBlockLocation: function(blockLocation) {
        this.prepareChange();
        
        // Set and notify listeners
        this.properties.blockLocation = blockLocation;
        this.fieldChanged("blockLocation", this.properties.blockLocation);
    },
    
    /**
     * setLockEnabled(enabled: boolean): void: Sets whether or not the lock is enabled for
     * this user
     *
     * Owner: Steven Lawrance
     */
    setLockEnabled: function(enabled) {
        this.prepareChange();
        
        // Set and notify listeners
        this.properties.lockEnabled = enabled;
        this.fieldChanged("lockEnabled", this.properties.lockEnabled);
    },
    
    /**
     * setPreventLockDisable(prevent: boolean): void: Set the registry permissions on keys
     * that are relevant to URL Lock to either prevent disabling of the URL Lock (true) or
     * allow disabling (false)
     *
     * Owner: Steven Lawrance
     */
    setPreventLockDisable: function(prevent) {
        this.prepareChange();
        
        // Set and notify listeners
        this.properties.preventLockDisable = prevent;
        this.fieldChanged("preventLockDisable", this.properties.preventLockDisable);
    },
    
    /**
     * setProcessAllLocations(blockAll: boolean): void: When set to true, all locations,
     * including drive letters, WebDAV shares, UNC paths, “My Computer,” “Control Panel,”
     * and other such locations, will be processed by the URL Lock. Note that this property
     * only affects the Internet Explorer version of the URL Lock and has no effect in the
     * Firefox version. When false, only web and FTP locations processed by the URL Lock
     *
     * Owner: Steven Lawrance
     */
    setProcessAllLocations: function(blockAll) {
        this.prepareChange();
        
        // Set and notify listeners
        this.properties.processAllLocations = blockAll;
        this.fieldChanged("processAllLocations", this.properties.processAllLocations);
    },
    
    /**
     * setShowDialogOnBlock(showDialog: boolean): void: When set to true, URL Lock will show
     * a dialog box when a site is blocked instead of a block page. When set to false, a
     * block page is shown when a site is blocked, which is defined by the blockLocation
     * property
     *
     * Owner: Steven Lawrance
     */
    setShowDialogOnBlock: function(showDialog) {
        this.prepareChange();
        
        // Set and notify listeners
        this.properties.showDialogOnBlock = blockAll;
        this.fieldChanged("processAllLocations", this.properties.showDialogOnBlock);
    },

    /**
     * Returns the requested interface for this object instance
     */
    QueryInterface: function(aIID)
    {
        if (!aIID.equals(Components.interfaces.IURLLockUser) &&
            !aIID.equals(nsISupports))
            throw Components.results.NS_ERROR_NO_INTERFACE;
        return this;
    }
}

/**
 * Creates a list array wrapper for LockUserList arrays
 */
function LockUserList(array) {
    this.array = array;
}

LockUserList.prototype = {

    item: function(index) {
        return this.array[index];
    },
    
    get length() {
        return this.array.length;
    }
}

/**
 * Creates a new lock rule instance using the given registry key
 */
function LockRule(configuration, key) {
    this.wrappedJSObject = this;

    // Set our configuration reference
    this.configuration = configuration;
    
    // Save our key as the rule ID
    this.ruleID = key;
    
    // Create our properties object
    this.properties = new Object();
    this.properties.name = key;
    
    // Create the users array to store the users that this rule applies to
    this.properties.users = new Object();
    this.properties.usersCached = null;
    
    // TODO: Load the regex and users from the key
    
    // Parse the regular expression, but don't notify listeners
    this.parseRegEx(false);

    // Set our oldProperties to null to indicate that we have no changes
    this.oldProperties = null;
}

LockRule.prototype = {    
    
    /**
     * addUser(user: LockUser): void: Adds the given user to this rule so that this rule
     * will apply to the given user
     *
     * Owner: Steven Lawrance
     */
    addUser: function(user) {
        this.prepareChange();
        
        // Add a user to the users array
        this.properties.users[user.getUserName()] = user;
        this.fieldChanged("users", this.properties.users);

        // Invalidate the user cache
        this.properties.usersCached = null;
    },
    
    /**
     * fieldChanged(fieldName: String, fieldValue: String): void: A field in this rule changed
     *
     * Owner: Steven Lawrance
     */
    fieldChanged: function(fieldName, fieldValue) {
    },
    
    /**
     * getDomain(): String: Returns the domain of this rule or, if unparseable, null
     * Owner: Steven Lawrance
     */
    getDomain: function() {
        return this.properties.domain;
    },
    
    /**
     * getName(): string: Returns the name of this rule
     */
    getName: function() {
        return this.properties.name;
    },
    
    /**
     * getOwner(): LockUser: Returns the user that created this regular expression.
     * If that user no longer exists, this gets set to the current user
     *
     * Owner: Steven Lawrance
     */
    getOwner: function() {
        return this.properties.owner;
    },
    
    /**
     * getLastModifiedTime(): date: Returns the date and time of the last modification to
     * this rule
     *
     * Owner: Steven Lawrance
     */
    getLastModifiedTime: function() {
        return this.properties.lastModifiedTime;
    },
    
    /**
     * getPathExceptions(): array<String>: Returns the path exceptions in this rule, which
     * are the paths that do not apply to this rule. If no exceptions are defined, this is
     * a blank array
     *
     * Owner: Steven Lawrance
     */
    getPathExceptions: function (count, list) {
        count.value = this.properties.pathExceptions.length;
        list.value = this.properties.pathExceptions;
    },
    
    /**
     * getPathList(list: array<String>): Returns the list of paths that apply to this rule.
     * If no paths apply to this rule, this is a blank array. If all paths apply to this
     * rule, this is null
     *
     * Owner: Steven Lawrance
     */
    getPathList: function(count, list) {
        count.value = this.properties.pathList.length;
        list.value = this.properties.pathList;
    },
    
    /**
     * getPortExceptions(): array<String>: Returns the port exceptions in this rule, which
     * are the ports that do not apply to this rule. If no exceptions are defined, this is
     * a blank array
     *
     * Owner: Steven Lawrance
     */
    getPortExceptions: function(count, list) {
        count.value = this.properties.portExceptions.length;
        list.value = this.properties.portExceptions;
    },
    
    /**
     * getPortList(): array<String>: Returns the list of ports that apply to this rule. If no
     * ports apply to this rule, this is a blank array. If all ports apply to this rule,
     * this is null
     *
     * Owner: Steven Lawrance
     */
    getPortList: function(count, list) {
        count.value = this.properties.portList.length;
        list.value = this.properties.portList;
    },
    
    /**
     * getProtocols(): LockProtocols: Returns the protocols that are applicable to this
     * rule. Note that direct modifications to the returned object have no effect until
     * setProtocols() is called with that object
     *
     * Owner: Steven Lawrance
     */
    getProtocols: function() {
        return this.properties.protocols;
    },
    
    /**
     * getRegEx(): String: Returns the regular expression of this rule, which is the only
     * property that is actually saved and includes information from the other properties
     * of this rule
     *
     * Owner: Steven Lawrance
     */
    getRegEx: function() {
        return this.properties.regex;
    },
    
    /**
     * Returns the unique ID of this rule
     */
   getRuleID: function() {
       return this.ruleID;
   },
    
    /**
     * getSubdomainExceptions(): array<String>: Returns the subdomain exceptions in this
     * rule, which are the subdomains that do not apply to this rule. If no exceptions are
     * defined, this is a blank array
     *
     * Owner: Steven Lawrance
     */
    getSubdomainExceptions: function(count, list) {
        count.value = this.properties.subdomainExceptions.length;
        list.value = this.properties.subdomainExceptions;
    },
    
    /**
     * getSubdomainList(): array<String>: Returns the subdomains that apply to this rule.
     * If no subdomains apply to this rule, this is a blank array. If all subdomains apply
     * to this rule, then this is null
     *
     * Owner: Steven Lawrance
     */
    getSubdomainList: function(count, list) {
        count.value = this.properties.subdomainList.length;
        list.value = this.properties.subdomainList;
    },
    
    /**
     * getUsers(): LockUser[]: Returns an array of users that this rule applies to. Do not
     * add and remove items from this array directly. Instead, use the addUser() and
     * removeUser() functions
     *
     * Owner: Steven Lawrance
     */
    getUsers: function() {
    
        // Return the cached copy if we have one
        if (this.properties.usersCached != null) {
            return this.properties.usersCached;
        }
        
        // Copy our active users into an array
        var arr = new Array();
        for (i in this.properties.users) {
            if (this.properties.users[i] != null) {
                arr.push(this.properties.users[i]);
            }
        }
        
        // Cache the read-only version of the user list
        this.properties.usersCached = new LockUserList(arr);
        
        // Return the user list
        return this.properties.usersCached;
    },
    
    /**
     * isChanged(): boolean: Returns true if at least one property in this rule has
     * changed since the last save or load
     *
     * Owner: Steven Lawrance
     */
    isChanged: function() {
    
        // If we have an old properties, then we are changed
        return this.oldProperties != null;
    },
    
    /**
     * isExpressionParseable(): boolean: Returns true if the regular expression was
     * successfully parsed into this object’s properties. If the regular expression
     * could not be parsed into this object’s properties, then the regular expression
     * will have its original value, but changes to the properties might erase the
     * original details of the unparseable regular expression. Note that this is separate
     * from whether or not a regular expression is valid, though invalid regular
     * expressions will also return false in this function
     *
     * Owner: Steven Lawrance
     */
    isExpressionParseable: function() {
        return this.properties.expressionParseable;
    },
    
    /**
     * isUserIncluded(): boolean: Returns true if the given user is included in
     * this rule or false if not
     */
    isUserIncluded: function(user) {
        return this.properties.users[user.getUserName()] == user;
    },
    
    /**
     * Parses the regular expression into properties and optionally notifies listeners
     */
    parseRegEx: function(notifyListeners) {
        // TODO: Implement
    },
    
    /**
     * Prepares to change the properties by copying the existing properties
     * into the old configuration object. This method is private
     */
    prepareChange: function() {
    
        // If we already have old properties, then return now
        if (this.oldProperties != null) {
            return;
        }
        
        // Create the old properties and copy them
        this.oldProperties = new Object();
        for (i in this.properties) {
        
            // For users, copy the users array one element at a time
            if (i == "users") {
                this.oldProperties.users = new Object();
                for (t in this.properties.users) {
                    this.oldProperties.users[t] = this.properties.users[t];
                }
            }
            
            // For protocols, copy the protocols object one element at a time
            else if (i == "protocols") {
                this.oldProperties.protocols = new Object();
                for (t in this.properties.protocols) {
                    this.oldProperties.protocols[t] = this.properties.protocols[t];
                }
            }
            
            // Copy other values directly
            else {
                this.oldProperties[i] = this.properties[i];
            }
        }
    },

    /**
     * removeUser(user: LockUser): void: Removes the given user from this rule so that
     * this rule will not apply to the given user
     *
     * Owner: Steven Lawrance
     */
    removeUser: function(user) {
        this.prepareChange();
        
        // Remove the given user from the array
        this.properties.users[user.getUserName()] = null;
        this.fieldChanged("users", this.properties.users);

        // Invalidate the user cache
        this.properties.usersCached = null;
    },
    
    /**
     * revert(): void Reverts any changes to the previously saved or loaded state,
     * canceling any changes since then. After this function completes,
     * isChanged() will return faslse
     */
    revert: function() {

        // Forget the current properties
        var revertedProperties = this.properties;
        this.properties = this.oldProperties;
        this.oldProperties = null;
        
        // Generate a change even for every property that differs
        for (i in this.properties) {
        
            // Always notify for users and protocols; these can
            // be compared one element at a time in a future version
            if (i == "users" || i == "protocols" ||
                this.properties[i] != revertedProperties[i]) {
                this.fieldChanged(i, this.properties[i]);
            }
        }
    },
    
    /**
     * save(): void: Saves any unsaved changes to this rule
     *
     * Owner: Steven Lawrance
     */
    save: function() {
        // TODO: Implement
        
        // If we don't have old properties, then there is nothing to save
        if (this.oldProperties == null) {
            return;
        }
        
        // Update the key used in the parent when the name changes
        var oldID = this.ruleID;
        this.ruleID = this.properties.name;
        
        // The old properties are no longer relevant
        this.oldProperties = null;
        
        // Update the rules list in our configuration unless if we are not
        // added yet
        if (this.configuration.getRule(oldID) != null) {
            this.configuration.updateRule(this, oldID);
        }
    },
    
    /**
     * setDomain(domain: String): void: Sets the domain of this rule
     * Owner: Steven Lawrance
     */
    setDomain: function(domain) {
        this.prepareChange();
        
        // Set the domain
        this.properties.domain = domain;
        this.updateRegEx();
        this.fieldChanged("domain", this.properties.domain);
    },
    
    /**
     * setName(name: string): void: Sets the name of this rule
     */
    setName: function(name) {
        this.prepareChange();
        
        // Set the name
        this.properties.name = name;
        this.fieldChanged("name", this.properties.name);
    },
    
    /**
     * setProtocols(protocols: LockProtocols): void: Sets the protocols that are
     * applicable to this rule, which may update the regular expression as a side-effect
     *
     * Owner: Steven Lawrance
     */
    setProtocols: function(protocols) {
        this.prepareChange();
        
        // Set the protocols
        this.properties.protocols = protocols;
        this.updateRegEx();
        this.fieldChanged("protocols", this.properties.protocols);
    },
    
    /**
     * setRegEx(regex: String): boolean: Sets the regular expression of this rule, which
     * is the only property that is actually saved and includes information from the other
     * properties of this rule. Note that when this function is called to set the regular
     * expression, the new expression is parsed and the other properties get updated to
     * their new values based on the given regular expression. This function returns true
     * if the given regular expression was parseable into the other properties, false if
     * it was not, or an exception if the regular expression library encountered an error
     * while parsing the given expression
     *
     * Owner: Steven Lawrance
     */
    setRegEx: function(regex) {
        this.prepareChange();
        
        // Set the regular expression
        this.properties.regex = regex;
        this.parseRegEx();
        this.fieldChanged("regex", this.properties.regex);
    },
    
    /**
     * setSubdomainExceptions(exceptions: array<String>): void: Sets the subdomain
     * exceptions in this rule, which are the subdomains that do not apply to this
     * rule. If no exceptions are to be defined, set this to a blank array
     *
     * Owner: Steven Lawrance
     */
    setSubdomainExceptions: function(count, list) {
        this.prepareChange();
        
        // Set the subdomain exceptions
        this.properties.subdomainExceptions = list;
        this.updateRegEx();
        this.fieldChanged("subdomainExceptions", this.properties.subdomainExceptions);
    },
    
    /**
     * setSubdomainList(list: array<String>): void: Sets the subdomains that apply to
     * this rule. If no subdomains are to be applied to this rule, set this to a blank
     * array. If all subdomains are to be applied to this rule, then set this to null
     *
     * Owner: Steven Lawrance
     */
    setSubdomainList: function(count, list){
        this.prepareChange();
        
        // Set the subdomain list
        this.properties.subdomainList = list;
        this.updateRegEx();
        this.fieldChanged("subdomainList", this.properties.subdomainList);
    },
    
    /**
     * setPortExceptions(exceptions: array<String>): void: Sets the port exceptions for
     * this rule, which are the ports that do not apply to this rule. If no exceptions
     * are defined, set this to a blank array
     *
     * Owner: Steven Lawrance
     */
    setPortExceptions: function(count, list) {
        this.prepareChange();
        
        // Set the port exceptions
        this.properties.portExceptions = list;
        this.updateRegEx();
        this.fieldChanged("portExceptions", this.properties.portExceptions);
    },
    
    /**
     * setPortList(list: array<String>): void: Sets the ports that apply to this rule. If
     * no ports are to be applied to this rule, set this to a blank array. If all ports
     * are to be applied to this rule, set this to null
     *
     * Owner: Steven Lawrance
     */
    setPortList: function(count, list) {
        this.prepareChange();
        
        // Set the port list
        this.properties.portList = list;
        this.updateRegEx();
        this.fieldChanged("portList", this.properties.portList);
    },
    
    /**
     * setPathExceptions(exceptions: array<String>): void: Sets the path exceptions for
     * this rule, which are the paths that do not apply to this rule. If no exceptions
     * are defined, set this to a blank array
     *
     * Owner: Steven Lawrance
     */
    setPathExceptions: function(count, list) {
        this.prepareChange();
        
        // Set the path exceptions
        this.properties.pathExceptions = list;
        this.updateRegEx();
        this.fieldChanged("pathExceptions", this.properties.pathExceptions);
    },
    
    /**
     * setPathList(paths: array<String>): void: Sets the paths that apply to this rule. If no
     * paths are to be applied to this rule, set this to a blank array. If all paths are
     * to be applied to this rule, set this to null
     *
     * Owner: Steven Lawrance
     */
    setPathList: function(count, list) {
        this.prepareChange();
        
        // Set the path list
        this.properties.pathList = list;
        this.updateRegEx();
        this.fieldChanged("pathList", this.properties.pathList);
    },
    
    /**
     * Updates the regular expression in response to a change in one or more fields
     */
    updateRegEx: function() {
        // TODO: Implement
    },

    /**
     * Returns the requested interface for this object instance
     */
    QueryInterface: function(aIID)
    {
        if (!aIID.equals(Components.interfaces.IURLLockRule) &&
            !aIID.equals(nsISupports))
            throw Components.results.NS_ERROR_NO_INTERFACE;
        return this;
    }
}

/**
 * List of rules to wrap a JavaScript array
 */
function LockRuleList(array) {
    this.array = array;
}

LockRuleList.prototype = {

    item: function(index) {
        return this.array[index];
    },
    
    get length() {
        return this.array.length;
    }
}

/**
 * Creates a new lock configuration object for the given engine
 */
function LockConfiguration(lockEngine) {
    this.wrappedJSObject = this;
    
    // Set our lock engine
    this.lockEngine = lockEngine;

    // Create our properties object
    this.properties = new Object();
    
    // Create the users array to store the users
    this.properties.users = new Object();
    this.properties.usersCached = null;
    
    // Create the users array to store the users
    this.properties.rules = new Object();
    this.properties.rulesCached = null;

    // Get the singleton native service    
    var lockNative = Components.classes["@ieurllock.sourceforge.net/native;1"]
                     .getService(Components.interfaces.IURLLockNative);

    // Load the list of users
    var count = new Object();
    var users = lockNative.getUserIDs(count);
    
    // Create a new LockUser for every user returned
    this.properties.users = new Object();
    var name;
    for (i = 0; i < count.value; i++) {
        name = lockNative.getUserName(users[i]);
        if (name != "ASPNET" && name != "Guest" && name != "HelpAssistant" &&
            name != "SQLDebugger" && name.indexOf("SUPPORT_") != 0)
        this.properties.users[name] = new LockUser(this, users[i]);
    }
    
    // Set the current user
    this.currentUser = this.properties.users[lockNative.getCurrentUser()];
    
    // Set the default user
    this.defaultUser = this.currentUser;  // TODO: Implement as an aggregation of all users
    
    // TODO: Load the rules
    
    // Set our oldProperties to null to indicate that we have no changes
    this.oldProperties = null;
}

LockConfiguration.prototype = {

    /**
     * addRule(rule: LockRule): void: Adds the given rule instance. This function
     * does not add the rule to the rule list; it only creates it. To add this rule to the
     * rule list, call addRule()
     *
     * Owner: Steven Lawrance
     */
    addRule: function(rule) {
        this.prepareChange();
        
        // Invalidate the rules cache
        this.properties.rulesCached = null;
        
        // Add the rule to the rules array
        this.properties.rules[rule.getRuleID()] = rule;
        this.fieldChanged("rules", this.properties.rules);
    },

    /**
     * createRule(): LockRule: Creates a new rule instance and returns it. This function
     * does not add the rule to the rule list; it only creates it. To add this rule to the
     * rule list, call addRule()
     *
     * Owner: Steven Lawrance
     */
    createRule: function() {
    
        // Create a new rule and return it
        return new LockRule(this, null);
    },
    
    /**
     * fieldChanged(fieldName: String, fieldValue: String): void: A field in the configuration
     * changed, which includes changes to users and rules
     *
     * Owner: Steven Lawrance
     */
    fieldChanged: function(fieldName, fieldValue) {
    },

    /**
     * getCurrentUser(): LockUser: Returns the current user, which is the user who is
     * running the configuration user interface. This can be used, as one example, to warn
     * the user if they are not an administrator
     *
     * Owner: Steven Lawrance
     */
    getCurrentUser: function() {
        return this.currentUser;
    },

    /**
     * getDefaultUser(): LockUser: Returns a virtual user that represents the system-wide
     * user. Properties in this virtual user are applied to the system-wide configuration
     *
     * Owner: Steven Lawrance
     */
    getDefaultUser: function() {
        return this.defaultUser;
    },

    /**
     * getRule(): LockRule: Returns the rule that has the given ID or null
     * if none have that ID
     */
    getRule: function(ruleID) {
        return this.properties.rules[ruleID];
    },

    /**
     * getRules(): LockRule[]: Get the rules that are defined in the system. Do not add
     * or remove elements from this array directly. Instead, use addRule() and removeRule()
     *
     * Owner: Steven Lawrance
     */
    getRules: function() {
    
        // Return the cached copy if we have one
        if (this.properties.rulesCached != null) {
            return this.properties.rulesCached;
        }
        
        // Copy our active rules into an array
        var arr = new Array();
        for (i in this.properties.rules) {
            if (this.properties.rules[i] != null) {
                arr.push(this.properties.rules[i]);
            }
        }
        
        // Cache the read-only version of the rule list
        this.properties.rulesCached = new LockRuleList(arr);
        
        // Return the user list
        return this.properties.rulesCached;
    },

    /**
     * getUser(): LockUser: Returns the user that has the given user name or null
     * if none have that user name
     */
    getUser: function(userName) {
        return this.properties.users[userName];
    },

    /**
     * getUsers(): LockUser[]: Get the users that are on the local computer. Do not add or
     * remove elements from this array
     *
     * Owner: Steven Lawrance
     */
    getUsers: function() {
    
        // Return the cached copy if we have one
        if (this.properties.usersCached != null) {
            return this.properties.usersCached;
        }
        
        // Copy our active users into an array
        var arr = new Array();
        for (i in this.properties.users) {
            if (this.properties.users[i] != null) {
                arr.push(this.properties.users[i]);
            }
        }
        
        // Cache the read-only version of the user list
        this.properties.usersCached = new LockUserList(arr);
        
        // Return the user list
        return this.properties.usersCached;
    },

    /**
     * isChanged(): boolean: Returns true if at least one property was changed in a user
     * or rule that had not been saved yet
     *
     * Owner: Steven Lawrance
     */
    isChanged: function() {
    
        // If we have an old properties, then we are changed
        if (this.oldProperties != null) {
            return true;
        }
        
        // If at least one user has changes, then we have unsaved changes
        for (i in this.properties.users) {
            if (this.properties.users[i] != null) {
                if (this.properties.users[i].isChanged()) {
                    return true;
                }
            }
        }

        // If at least one rule has changes, then we have unsaved changes
        for (i in this.properties.rules) {
            if (this.properties.rules[i] != null) {
                if (this.properties.rules[i].isChanged()) {
                    return true;
                }
            }
        }
        
        // We are not changed if we made it this far
        return false;
    },

    /**
     * Prepares to change the properties by copying the existing properties
     * into the old configuration object. This method is private
     */
    prepareChange: function() {
    
        // If we already have old properties, then return now
        if (this.oldProperties != null) {
            return;
        }
        
        // Create the old properties and copy them
        this.oldProperties = new Object();
        for (i in this.properties) {
            this.oldProperties[i] = this.properties[i];
        }
    },
    
    /**
     * removeRule(rule: LockRule): void: Removes the given rule object from the list of
     * defined rules
     *
     * Owner: Steven Lawrance
     */
    removeRule: function(rule) {
        this.prepareChange();
        
        // Invalidate the rules cache
        this.properties.rulesCached = null;
        
        // Remove the rule from the rules array
        this.properties.rules[rule.getRuleID()] = null;
        this.fieldChanged("rules", this.properties.rules);
    },
    
    /**
     * save(): void: Saves any changes to the users and rules in this configuration
     *
     * Owner: Steven Lawrance
     */
    save: function() {
        // TODO: Implement
        
        // Save all users and rules

        // If we don't have old properties, then there is nothing to save
        if (this.oldProperties == null) {
            return;
        }
        
        // The old properties are no longer relevant
        this.oldProperties = null;
    },

    /**
     * Updates the given user and notifies listeners
     */
    updateUser: function(user) {
        
        // Notify listeners that the users list changed
        this.fieldChanged("users", this.properties.users);
    },

    /**
     * Updates the given rule, adjusting the map if the name changed
     * and notifying listeners that the list changed
     */
    updateRule: function(rule, oldID) {
        
        // Invalidate the rules cache
        this.properties.rulesCached = null;
        
        // If the old ID does not equal the new ID, then
        // adjust the map
        if (rule.getRuleID() != oldID) {
            this.properties.rules[oldID] = null;
            this.properties.rules[rule.getRuleID()] = rule;
        }
        
        // Notify listeners that the rules list changed
        this.fieldChanged("rules", this.properties.rules);
    },

    /**
     * Returns the requested interface for this object instance
     */
    QueryInterface: function(aIID)
    {
        if (!aIID.equals(Components.interfaces.IURLLockConfiguration) &&    
            !aIID.equals(nsISupports))
            throw Components.results.NS_ERROR_NO_INTERFACE;
        return this;
    }
};

function LockEngine() {
    this.wrappedJSObject = this;

    // Lazily-created configuration instance
    this.configuration = null;
}

LockEngine.prototype = {
    
    /**
     * getConfiguration(): LockConfiguration: Returns the LockConfiguration object,
     * which provides access to the configuration and to the users
     *
     * Owner: Steven Lawrance
     */
    getConfiguration: function() {
        
        // Create the configuration if we need to
        if (this.configuration == null) {
            this.configuration = new LockConfiguration(this);
        }
        
        // Return our configuration
        return this.configuration;
    },
    
    /**
     * isLocationBlocked(location: String): boolean: Returns true if the given location
     * or URL is blocked or false if it is not blocked
     *
     * Owner: Steven Lawrance
     */
    isLocationBlocked: function(location) {
		
		// Loop through all the rules (not regular expressions yet)
		var rules = this.getConfiguration().getRules();
		var rule;
		var currentUser = this.getConfiguration().getCurrentUser();
		for (i = 0; i < rules.length; i++) {
			rule = rules.item(i);
			
			// If the current user is in this rule and it's not chrome://, then proceed
			if (rule.isUserIncluded(currentUser) && location.indexOf("chrome://") != 0) {

				// If the given location contains the rule, then we are blocked
				// TODO: Use PCRE and regular expressions in a future version like
				// the IE URL Lock
				if (location.indexOf(rule.getDomain()) >= 0) {
				    return true;
				}
			}
		}
		
        // We are not blocked
        return false;
    },

    /**
     * Returns the requested interface for this object instance
     */
    QueryInterface: function(aIID)
    {
        if (!aIID.equals(Components.interfaces.IURLLockEngine) &&
            !aIID.equals(nsISupports))
            throw Components.results.NS_ERROR_NO_INTERFACE;
        return this;
    }
}

/**
 * Factory object for LockEngine. In JavaScript, you
 * can get the existing instance with the following code:
 * myConfiguration = Components.classes["@ieurllock.sourceforge.net/engine;1"].
 *                              createInstance(Components.interfaces.IURLLockEngine);
 */
var LockEngineFactory = {
    createInstance: function (aOuter, aIID)
    {
        if (aOuter != null)
            throw Components.results.NS_ERROR_NO_AGGREGATION;
        return (new LockEngine()).QueryInterface(aIID);
    }
};

var LockEngineModule = {
    _firstTime: true,
    registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
    {
        if (this._firstTime) {
            this._firstTime = false;
            throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
        };
        aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
        aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME, CONTRACT_ID, aFileSpec,
                                         aLocation, aType);
    },

    unregisterSelf: function(aCompMgr, aLocation, aType)
    {
        aCompMgr = aCompMgr.
            QueryInterface(Components.interfaces.nsIComponentRegistrar);
        aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);        
    },
  
    getClassObject: function(aCompMgr, aCID, aIID)
    {
        if (!aIID.equals(Components.interfaces.nsIFactory))
            throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
        if (aCID.equals(CLASS_ID))
            return LockEngineFactory;
        throw Components.results.NS_ERROR_NO_INTERFACE;
    },

    canUnload: function(aCompMgr) { return true; }
};

/***********************************************************
module initialization

When the application registers the component, this function
is called.
***********************************************************/
function NSGetModule(aCompMgr, aFileSpec) { return LockEngineModule; }
